import re
from typing import Tuple, List, Dict

EOF = -1
SKIP = -2

ID = 0
STR_ID = 1
HTML_ID = 2
EDGE_OP = 3

LSQUARE = 4
RSQUARE = 5
LCURLY = 6
RCURLY = 7
COMMA = 8
COLON = 9
SEMI = 10
EQUAL = 11
PLUS = 12

STRICT = 13
GRAPH = 14
DIGRAPH = 15
NODE = 16
EDGE = 17
SUBGRAPH = 18


class Scanner:
  """Stateless scanner."""

  # should be overriden by derived classes
  tokens: List[Tuple[int, bytes, bool]] = []
  symbols: Dict[bytes, int] = {}
  literals: Dict[bytes, int] = {}
  ignorecase = False

  def __init__(self) -> None:
    flags = re.DOTALL
    if self.ignorecase:
      flags |= re.IGNORECASE
    self.tokens_re = re.compile(
        b'|'.join([b'(' + regexp + b')'
                   for type, regexp, test_lit in self.tokens]),
        flags
    )

  def next(self, buf: bytes, pos: int) -> Tuple[int, bytes, int]:
    if pos >= len(buf):
      return EOF, b'', pos
    mo = self.tokens_re.match(buf, pos)
    if mo:
      text = mo.group()
      type, regexp, test_lit = self.tokens[mo.lastindex - 1]
      pos = mo.end()
      if test_lit:
        type = self.literals.get(text, type)
      return type, text, pos
    else:
      c = buf[pos:pos + 1]
      return self.symbols.get(c, None), c, pos + 1


class DotScanner(Scanner):
  # token regular expression table
  tokens = [
    # whitespace and comments
    (SKIP,
     br'[ \t\f\r\n\v]+|'
     br'//[^\r\n]*|'
     br'/\*.*?\*/|'
     br'#[^\r\n]*',
     False),

    # Alphanumeric IDs
    (ID, br'[a-zA-Z_\x80-\xff][a-zA-Z0-9_\x80-\xff]*', True),

    # Numeric IDs
    (ID, br'-?(?:\.[0-9]+|[0-9]+(?:\.[0-9]*)?)', False),

    # String IDs
    (STR_ID, br'"[^"\\]*(?:\\.[^"\\]*)*"', False),

    # HTML IDs
    (HTML_ID, br'<[^<>]*(?:<[^<>]*>[^<>]*)*>', False),

    # Edge operators
    (EDGE_OP, br'-[>-]', False),
  ]

  # symbol table
  symbols = {
    b'[': LSQUARE,
    b']': RSQUARE,
    b'{': LCURLY,
    b'}': RCURLY,
    b',': COMMA,
    b':': COLON,
    b';': SEMI,
    b'=': EQUAL,
    b'+': PLUS,
  }

  # literal table
  literals = {
    b'strict': STRICT,
    b'graph': GRAPH,
    b'digraph': DIGRAPH,
    b'node': NODE,
    b'edge': EDGE,
    b'subgraph': SUBGRAPH,
  }

  ignorecase = True
